/* audio.c - interface to system audio
 *
 * This version uses PortAudio (http://www.portaudio.com).
 *
 * Copyright 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "audio.h"
#include <glib.h>
#include <portaudio.h>

const char *last_audio_error;

/* Whether PortAudio has been initialized or not. */
static int initialized = 0;

/* Available output devices. */
static Audio_device *devices = NULL;
static int ndevices;

int
audio_init(void)
{
    if (initialized) return 0;
    g_debug("initializing PortAudio");
    PaError rv = Pa_Initialize();
    if (rv != paNoError) {
        last_audio_error = Pa_GetErrorText(rv);
        return 1;
    } else {
        initialized = 1;
        return 0;
    }
}

int audio_close(void)
{
    if (!initialized) return 0;
    g_debug("terminating PortAudio");
    if (devices) {
        g_free(devices);
        devices = NULL;
    }
    PaError rv = Pa_Terminate();
    if (rv != paNoError) {
        last_audio_error = Pa_GetErrorText(rv);
        return 1;
    } else {
        initialized = 0;
        return 0;
    }
}

int
audio_get_outputs(const Audio_device **devs)
{
    g_assert(initialized);
    PaDeviceIndex count = Pa_GetDeviceCount();
    if (count < 0) {
        last_audio_error = Pa_GetErrorText(count);
        return -1;
    }
    if (!devices) {
        devices = g_malloc(count * sizeof(Audio_device));
        ndevices = count;
    } else if (count != ndevices) {
        devices = g_realloc(devices, count * sizeof(Audio_device));
        ndevices = count;
    }
    PaDeviceIndex default_dev = Pa_GetDefaultOutputDevice();
    for (PaDeviceIndex i = 0; i < ndevices; i++) {
        const PaDeviceInfo *devinfo = Pa_GetDeviceInfo(i);
        if (!devinfo) {
            g_warning("failed to fetch device information "
                      "for device %d", i);
            continue;
        }
        devices[i].index = i;
        devices[i].name = devinfo->name;
        devices[i].latency = 1000 * devinfo->defaultHighOutputLatency + 0.5;
        devices[i].sysdefault = (i == default_dev);
    }
    *devs = devices;
    return ndevices;
}
