/* gtkui.c - GTK GUI implementation
 *
 * Copyright 2010, 2013, 2016, 2024, 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gtkui.h"
#include "audio.h"
#include "config.h"
#include "playback.h"
#include "test.h"
#include <stdio.h>
#include <stdlib.h>

/* Current test trial (zero-based), or -1 if there is no test in
 * progress. */
int current_trial;

/* List of icon pixmaps. */
GList *icons;

/* Default number of test trials. */
static gint ntrials = 16;
/* Default output device (use PortAudio default). */
static gint outdev = -1;
/* Default latency (use PortAudio default). */
static gint latency = -1;
/* Should we scan and list audio devices? */
static gboolean listdevs = FALSE;

/*
 * Parse command line options.
 *
 * Exit the program with EXIT_FAILURE in case of errors.
 */
static void
parse_options(int *argc, char **argv[])
{
    GOptionEntry entries[] = {
        { "num-trials", 'i', 0, G_OPTION_ARG_INT, &ntrials,
          "Do N trials", "N" },
        { "audio-device", 'd', 0, G_OPTION_ARG_INT, &outdev,
          "Use PortAudio audio device ID", "ID" },
        { "list-devices", 'D', 0, G_OPTION_ARG_NONE, &listdevs,
          "List PortAudio devices and exit", NULL },
        { "latency", 'l', 0, G_OPTION_ARG_INT, &latency,
          "Set suggested output latency to L milliseconds", "L" },
        G_OPTION_ENTRY_NULL
    };

    GOptionContext *context = g_option_context_new("[SAMPLE-A SAMPLE-B]");
    g_option_context_set_summary
        (context, "Test for audible differences in fidelity between "
         "two audio samples\nwith ABX test.");
    g_option_context_set_description
        (context, "Report bugs to " PACKAGE_BUGREPORT ".");
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gtk_get_option_group(TRUE));
    GError *error = NULL;
    if (!g_option_context_parse(context, argc, argv, &error)) {
        fprintf(stderr, "Error: %s\n", error->message);
        g_option_context_free(context);
        exit(EXIT_FAILURE);
    }
    g_option_context_free(context);

    if (ntrials <= 0) {
        fprintf(stderr, "Error: the number of trials must be "
                "greater than zero\n");
        exit(EXIT_FAILURE);
    }
}

/*
 * Show an error dialog for failed sample playback initialization.
 */
static
void show_sample_playback_init_error(const gchar *sample,
                                     GtkWidget *parent)
{
    gchar *basename = g_path_get_basename(sample);
    gchar *error = g_strdup_printf
        ("Cannot set up playback for sample %s.\n\n%s",
         basename, playback_error);
    error_dialog(parent, "Failed to set up a test", error);
    g_free(basename);
    g_free(error);
}

/*
 * Create pixmaps from window icon resources.
 */
static void
create_icons()
{
    g_assert(icons == NULL);
    const char *paths[]  = {"/fi/kapsi/phintsan/abx/abx-16.png",
                            "/fi/kapsi/phintsan/abx/abx-32.png",
                            "/fi/kapsi/phintsan/abx/abx-48.png",
                            "/fi/kapsi/phintsan/abx/abx-64.png",
                            "/fi/kapsi/phintsan/abx/abx-80.png",
                            "/fi/kapsi/phintsan/abx/abx-96.png",
                            "/fi/kapsi/phintsan/abx/abx-128.png",
                            "/fi/kapsi/phintsan/abx/abx-160.png"};
    size_t npaths = sizeof(paths) / sizeof(paths[0]);
    for (int i = 0; i < npaths; i++) {
        GError *err = NULL;
        GdkPixbuf* icon = gdk_pixbuf_new_from_resource(paths[i], &err);
        if (!icon) {
            g_error("failed to read pixmap resource %s", paths[i]);
        }
        icons = g_list_append(icons, icon);
    }
}

/*
 * Initialize and pass control to GTK.
 *
 * Initialize GTK, the audio subsystem, and a new test (if specified
 * on the command line).  After GTK is finished, clean up resources
 * and return EXIT_SUCCESS.  In case of error return EXIT_FAILURE.
 */
int
gtk_ui_main(int argc, char *argv[])
{
    int rv;
    parse_options(&argc, &argv);
    gtk_init(&argc, &argv);
    create_icons();
    if (audio_init() != 0) {
        error_dialog(NULL, "Failed to initialize the audio subsystem",
                     last_audio_error);
        rv = EXIT_FAILURE;
        goto out;
    }
    if (listdevs) {
        const Audio_device *devices;
        int ndevices = audio_get_outputs(&devices);
        if (ndevices == -1) {
            printf("No output devices found.\n");
            rv = EXIT_FAILURE;
            goto out;
        }
        printf("Output devices:\n");
        for (int i = 0; i < ndevices; i++) {
            printf("Device %2d: '%s', latency %d ms",
                   devices[i].index,
                   devices[i].name,
                   devices[i].latency);
            if (devices[i].sysdefault) printf (" [default]\n");
            else printf("\n");
        }
        rv = EXIT_SUCCESS;
        goto out;
    }
    if (playback_init(outdev, latency) != 0) {
        error_dialog(NULL, "Failed to set up playback",
                     playback_error);
        rv = EXIT_FAILURE;
        goto out;
    }
    if (argc == 2) {
        error_dialog(NULL, "Failed to set up a test",
                     "Only one sample filename given in the command line.\n"
                     "Two samples are required.");
        rv = EXIT_FAILURE;
        goto out;
    } else if (argc >= 3) {
        if (set_up_test(argv[1], argv[2], ntrials, NULL) == 0)
        {
            /* test initialization successful */
            show_main_window();
            reset_test_widgets();
        } else {
            rv = EXIT_FAILURE;
            goto out;
        }
    } else {
        /* no current test available */
        current_trial = -1;
        show_main_window();
        update_main_window();
    }
    gtk_main();
    rv = EXIT_SUCCESS;

out:
    /* Clean up. */
    discard_test();
    playback_close();
    if (audio_close() != 0) {
        g_warning("failed to shut down audio subsystem: %s",
                  playback_error);
    }
    return rv;
}

/*
 * Display a simple error dialog with the given parent window and
 * error message.
 */
void
error_dialog(GtkWidget *parent_win, const gchar *title,
             const gchar *message)
{
    GtkWidget *errdialog;
    errdialog = gtk_message_dialog_new
        (GTK_WINDOW(parent_win), GTK_DIALOG_DESTROY_WITH_PARENT,
         GTK_MESSAGE_ERROR, GTK_BUTTONS_CLOSE, "%s", title);
    gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(errdialog),
                                             "%s", message);
    gtk_window_set_title(GTK_WINDOW(errdialog), "Error");
    gtk_dialog_run(GTK_DIALOG(errdialog));
    gtk_widget_destroy(errdialog);
}

/*
 * Display a simple warning dialog with the given parent window and
 * warning message.
 */
void
warning_dialog(GtkWidget *parent_win, const gchar *title,
               const gchar *message)
{
    GtkWidget *dialog;
    dialog = gtk_message_dialog_new
        (GTK_WINDOW(parent_win), GTK_DIALOG_DESTROY_WITH_PARENT,
         GTK_MESSAGE_WARNING, GTK_BUTTONS_CLOSE, "%s", title);
    gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
                                             "%s", message);
    gtk_window_set_title(GTK_WINDOW(dialog), "Warning");
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

/*
 * Set up a new test and playback for the given samples.  In case of
 * errors, show an error dialog with the given parent window.
 *
 * Return zero on success, nonzero otherwise.
 */
int set_up_test(const gchar *a, const gchar *b,
                int ntrials, GtkWidget *parent)
{
    /* Discard the current test if one is in progress. */
    discard_test();

    /* Initialize playback for the samples. */
    int res = playback_open_samples(a, b);
    if (res != 0) {
        switch (res) {
        case 1:
        case 2:
        {
            const gchar *sample = (res == 1 ? a : b);
            show_sample_playback_init_error(sample, parent);
            return res;
        }
        case 3:
            warning_dialog(parent, "Difference in sample durations",
                           "The chosen samples have different durations. Both samples\n"
                           "should have approximately the same duration.\n\n"
                           "Continue the test only if you know what you are doing.");
            break;
        case 4:
        case 5:
        {
            const gchar *sample = (res == 4 ? a : b);
            show_sample_playback_init_error(sample, parent);
            return res;
        }
        }
    }

    /* Initialize a new test. */
    if (test_init(ntrials) != 0) {
        error_dialog(parent, "Failed to set up a test",
                     "Invalid number of trials.");
        playback_close_samples();
        return 6;
    }

    return 0;
}

/*
 * Discard the current test.
 */
void
discard_test(void)
{
    playback_close_samples();
    test_close();
}
