/* mainwin.c - main window
 *
 * Copyright 2010, 2013, 2016, 2024, 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "config.h"
#include "gtkui.h"
#include "playback.h"
#include "soundfile.h"
#include "test.h"
#include <math.h>

static Sndfile_metadata metadata_a; /* Metadata for sample A. */
static Sndfile_metadata metadata_b; /* Metadata for sample B. */
static gchar *basename_a;           /* Basename for file A. */
static gchar *basename_b;           /* Basename for file B. */

static GtkWidget *main_window;

/* How often we should update the location scale (in ms).  Note that
   GTK can't handle very rapid updates. */
static const int UPDATE_LOCATION_TIMEOUT = 1000; 

/* Delay for doing an actual seek operation after a scrolling event
 * happens on the location scale.  This is used to prevent bursts of
 * (redundant) seek requests, for example when the user rolls the
 * scroll wheel. */
static const int SEEK_DELAY_TIMEOUT = 250;

/* Glib event source id for timers. */
static guint update_timeout_id;
static guint seek_delay_id;

static GtkWidget *new_test_menu_item;
static GtkWidget *about_menu_item;
static GtkWidget *quit_menu_item;

static GtkWidget *status_bar;
static gint play_context;
static gint non_play_context;

static GtkWidget *first_sample;
static GtkWidget *second_sample;
static GtkWidget *pause_button;
static GtkWidget *test_sample;
static GtkWidget *prev_marker;
static GtkWidget *next_marker;

static GtkAdjustment *adjustment;
static GtkWidget *scale;
static int is_user_seeking;

static GtkWidget *markers;
static GtkListStore *marker_list;

static GtkWidget *trial_label;
static GtkWidget *meta_a_label;
static GtkWidget *meta_b_label;

static gboolean recurrent_update_location(gpointer data);
static void delayed_seek(gpointer data);
static void clear_status_bar(void);

static GtkWidget *create_metadata_box(void);
static GtkWidget *create_test_box(void);
static GtkWidget *create_menu_bar(void);
static void create_main_window(void);
static void create_status_bar(void);
static GtkWidget *create_icon_button(const gchar *icon, gchar *label);
static GtkWidget *create_playback_box(void);
static GtkWidget *create_marker_box(void);
static void format_marker(GtkTreeViewColumn *tree_column,
                          GtkCellRenderer *cell,
                          GtkTreeModel *tree_model,
                          GtkTreeIter *iter,
                          gpointer data);

static gboolean delete_event_handler(GtkWidget *widget,
                                     gpointer data);
static void destroy_event_handler(GtkWidget *widget,
                                  gpointer data);
static void marker_activated(GtkTreeView *markers, 
                             GtkTreePath *path,
                             GtkTreeViewColumn *column, 
                             gpointer user_data);
static void clear_marker_selection(void);

static void rewind_button_clicked(GtkWidget *widget, gpointer data);
static void pause_button_clicked(GtkWidget *widget, gpointer data);
static void play_button_clicked(GtkWidget *widget, gpointer data);
static void decide_button_clicked(GtkWidget *widget, gpointer data);
static gboolean scale_button_pressed_or_released(GtkWidget *widget, 
                                                 GdkEventButton *event,
                                                 gpointer data);
static gboolean scale_scroll_action(GtkRange *range,
                                    GtkScrollType scroll,
                                    double value,
                                    gpointer data);
static void add_marker_clicked(GtkWidget *widget, gpointer data);
static void remove_marker_clicked(GtkWidget *widget, gpointer data);
static void menu_item_clicked(GtkWidget *widget, gpointer data);

static gboolean confirm(const gchar *message);
static void show_about_dialog(void);
static void decide_current_trial(void);

/*
 * Show the main window, creating it if necessary.
 */
void
show_main_window(void)
{
    if (!main_window) {
        create_main_window();
        gtk_widget_show_all(main_window);
    } else {
        gtk_window_present(GTK_WINDOW(main_window));
    }
}

/*
 * Update the main window contents to the current test progress.
 */
void
update_main_window(void)
{
    static GString *label = NULL;
    static GString *meta_a = NULL;
    static GString *meta_b = NULL;

    if (!label) label = g_string_new(NULL);
    if (!meta_a) meta_a = g_string_new(NULL);
    if (!meta_b) meta_b = g_string_new(NULL);

    if (current_trial < 0) {
        g_string_printf(label, "No test in progress.");
        g_string_printf(meta_a, "(none)");
        g_string_printf(meta_b, "(none)");
        g_object_set(adjustment, "upper", 0.0, "value", 0.0, NULL);
        gtk_list_store_clear(marker_list);
    } else if (current_trial == 0) {
        double limit = MIN(metadata_a.duration, metadata_b.duration);
        g_string_printf(label, "Test trial 1 of %d", test_get_ntrials());
        g_string_printf(meta_a, 
                        "%s, %d Hz, %d channels",
                        basename_a,
                        metadata_a.rate,
                        metadata_a.channels);
        g_string_printf(meta_b,
                        "%s, %d Hz, %d channels",
                        basename_b,
                        metadata_b.rate,
                        metadata_b.channels);
        g_object_set(adjustment, "upper", (gdouble) limit,
                     "value", 0.0, NULL);
        gtk_list_store_clear(marker_list);
    } else {
        g_string_printf(label, "Test trial %d of %d",
                        current_trial + 1, test_get_ntrials());
    }

    gtk_label_set_text(GTK_LABEL(trial_label), label->str);
    gtk_label_set_text(GTK_LABEL(meta_a_label), meta_a->str);
    gtk_label_set_text(GTK_LABEL(meta_b_label), meta_b->str);
}

/*
 * Reset test parameters in the window.  This function is usually
 * called after a new test has been successfully initialized.
 */
void
reset_test_widgets(void)
{
    if (basename_a) g_free(basename_a);
    if (basename_b) g_free(basename_b);
    current_trial = 0;
    playback_get_metadata(&metadata_a, &metadata_b);
    basename_a = g_path_get_basename(metadata_a.filename);
    basename_b = g_path_get_basename(metadata_b.filename);
    update_main_window();
}

/*
 * Update time slider to the current playback location.  Remove
 * possible previous recurrent update timer, and reinstall a new one.
 * (This is done mostly for consistent UI behavior.)
 */
static void
update_location(void)
{
    if (update_timeout_id) {
        g_source_remove(update_timeout_id);
        update_timeout_id = 0;
    }
    if (recurrent_update_location(NULL)) {
        update_timeout_id = g_timeout_add(UPDATE_LOCATION_TIMEOUT,
                                          recurrent_update_location, NULL);
    }
}

/*
 * Update current location to the time slider unless the user is
 * seeking.  This function is added to the Glib main event loop
 * whenever update_location is called.
 *
 * Return FALSE if playback has been stopped, so that glib will remove
 * the function from its event loop.  Otherwise, return TRUE.
 */
static gboolean
recurrent_update_location(gpointer data)
{
    (void)(data);
    Playback_state state = playback_get_state();
    if (!is_user_seeking && !seek_delay_id && state.location >= 0) {
        gtk_range_set_value(GTK_RANGE(scale), state.location);
    }
    switch (state.playback) {
    case STOPPED:
        clear_status_bar();
        update_timeout_id = 0;
        return FALSE;
    case PAUSED:
        update_timeout_id = 0;
        return FALSE;
    case PLAYING:
        return TRUE;
    }
    return TRUE;
}

/*
 * Seek to the current scale position.  This function is meant to be
 * called by the seek delay timer.
 */
static void
delayed_seek(gpointer data)
{
    (void)(data);
    playback_seek(gtk_range_get_value(GTK_RANGE(scale)));
    update_location();
    seek_delay_id = 0;
}

/*
 * Restore "Stopped" to the status bar.
 */
static void
clear_status_bar(void)
{
    /* pop twice to remove possible "Paused" from the status bar */
    gtk_statusbar_pop(GTK_STATUSBAR(status_bar), play_context);
    gtk_statusbar_pop(GTK_STATUSBAR(status_bar), play_context);
}

/*
 * Create main window elements and lay them into main_box.
 */
static void
create_main_window(void)
{
    main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    GtkWidget *playback_box = create_playback_box();
    GtkWidget *test_box = create_test_box();
    GtkWidget *marker_box = create_marker_box();
    GtkWidget *meta_box = create_metadata_box();
    GtkWidget *menu_bar = create_menu_bar();
    create_status_bar();

    GtkWidget *left_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 20);
    gtk_box_pack_start(GTK_BOX(left_box), test_box, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(left_box), meta_box, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(left_box), playback_box, TRUE, FALSE, 0);

    GtkWidget *main_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 20);
    gtk_container_set_border_width(GTK_CONTAINER(main_hbox), 10);
    gtk_box_pack_start(GTK_BOX(main_hbox), left_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_hbox), marker_box, TRUE, TRUE, 0);

    GtkWidget *main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
    gtk_box_pack_start(GTK_BOX(main_box), menu_bar, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), main_hbox, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), status_bar, FALSE, FALSE, 0);

    gtk_window_set_title(GTK_WINDOW(main_window), "ABX Tester");
    gtk_window_set_icon_list(GTK_WINDOW(main_window), icons);
    g_signal_connect(G_OBJECT(main_window), "delete-event",
                     G_CALLBACK(delete_event_handler), NULL);
    g_signal_connect(G_OBJECT(main_window), "destroy",
                     G_CALLBACK(destroy_event_handler), NULL);
    gtk_container_add(GTK_CONTAINER(main_window), main_box);
}

/*
 * Return a new button with a stock icon.
 */
static GtkWidget *
create_icon_button(const gchar *icon, gchar *label)
{
    GtkWidget *button = gtk_button_new();
    GtkWidget *aligned_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
    GtkWidget *box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
    gtk_box_set_spacing(GTK_BOX(box), 5);
    gtk_container_add(GTK_CONTAINER(box),
                      gtk_image_new_from_icon_name(icon, GTK_ICON_SIZE_BUTTON));
    gtk_container_add(GTK_CONTAINER(box),
                      gtk_label_new_with_mnemonic(label));
    gtk_box_pack_start(GTK_BOX(aligned_box), box, TRUE, FALSE, 0);
    gtk_container_add(GTK_CONTAINER(button), aligned_box);
    return button;
}

/*
 * Create playback-related widgets (time scale and control buttons),
 * and lay them into a vbox.  Return the box.
 */
static GtkWidget *
create_playback_box(void)
{
    adjustment = gtk_adjustment_new(0, 0, 60, 1.0, 10.0, 0);
    scale = gtk_scale_new(GTK_ORIENTATION_HORIZONTAL, adjustment);
    gtk_scale_set_digits(GTK_SCALE(scale), 1);
    is_user_seeking = 0;
    g_signal_connect(G_OBJECT(scale), "button_press_event",
                     G_CALLBACK(scale_button_pressed_or_released), NULL);
    g_signal_connect(G_OBJECT(scale), "button_release_event",
                     G_CALLBACK(scale_button_pressed_or_released), NULL);
    g_signal_connect(G_OBJECT(scale), "change_value",
                     G_CALLBACK(scale_scroll_action), NULL);

    first_sample = create_icon_button("media-playback-start", "_A");
    g_signal_connect(G_OBJECT(first_sample), "clicked",
                     G_CALLBACK(play_button_clicked), NULL);
    second_sample = create_icon_button("media-playback-start", "_B");
    g_signal_connect(G_OBJECT(second_sample), "clicked",
                     G_CALLBACK(play_button_clicked), NULL);
    test_sample = create_icon_button("media-playback-start", "_X");
    g_signal_connect(G_OBJECT(test_sample), "clicked",
                     G_CALLBACK(play_button_clicked), NULL);
    pause_button = create_icon_button("media-playback-pause", "_Pause");
    g_signal_connect(G_OBJECT(pause_button), "clicked",
                     G_CALLBACK(pause_button_clicked), NULL);
    prev_marker = create_icon_button("media-skip-backward", "P_revious");
    g_signal_connect(G_OBJECT(prev_marker), "clicked",
                     G_CALLBACK(rewind_button_clicked), NULL);
    next_marker = create_icon_button("media-skip-forward", "_Next");
    g_signal_connect(G_OBJECT(next_marker), "clicked",
                     G_CALLBACK(rewind_button_clicked), NULL);

    GtkWidget *button_grid = gtk_grid_new();
    gtk_grid_set_column_homogeneous(GTK_GRID(button_grid), TRUE);
    gtk_widget_set_halign(button_grid, GTK_ALIGN_CENTER);
    gtk_widget_set_valign(button_grid, GTK_ALIGN_CENTER);

    gtk_grid_attach(GTK_GRID(button_grid), first_sample, 0, 0, 1, 1);
    gtk_grid_attach(GTK_GRID(button_grid), second_sample, 1, 0, 1, 1);
    gtk_grid_attach(GTK_GRID(button_grid), test_sample, 2, 0, 1, 1);
    gtk_grid_attach(GTK_GRID(button_grid), pause_button, 0, 1, 1, 1);
    gtk_grid_attach(GTK_GRID(button_grid), prev_marker, 1, 1, 1, 1);
    gtk_grid_attach(GTK_GRID(button_grid), next_marker, 2, 1, 1, 1);
    gtk_grid_set_row_spacing(GTK_GRID(button_grid), 10);
    gtk_grid_set_column_spacing(GTK_GRID(button_grid), 10);

    GtkWidget *playback_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 20);
    gtk_box_pack_start(GTK_BOX(playback_box), scale, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(playback_box), button_grid, FALSE, FALSE, 0);
    return playback_box;
}

/*
 * Create marker-related widgets and lay them into a vbox.  Return the
 * box.
 */
static GtkWidget *
create_marker_box(void)
{
    marker_list = gtk_list_store_new(1, G_TYPE_DOUBLE);
    markers = gtk_tree_view_new_with_model(GTK_TREE_MODEL(marker_list));
    g_signal_connect(G_OBJECT(markers), "row-activated",
                     G_CALLBACK(marker_activated),  NULL);
    GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes
        ("_Markers", renderer, "text", 0, NULL);
    gtk_tree_view_column_set_cell_data_func(column, renderer, format_marker,
                                            NULL, NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(markers), column);

    GtkWidget *add_marker = gtk_button_new_from_icon_name
        ("list-add", GTK_ICON_SIZE_BUTTON);
    GtkWidget *remove_marker = gtk_button_new_from_icon_name
        ("list-remove", GTK_ICON_SIZE_BUTTON);
    GtkWidget *clear_markers = gtk_button_new_from_icon_name
        ("edit-clear", GTK_ICON_SIZE_BUTTON);
    g_signal_connect(G_OBJECT(add_marker), "clicked",
                     G_CALLBACK(add_marker_clicked),  NULL);
    g_signal_connect(G_OBJECT(remove_marker), "clicked",
                     G_CALLBACK(remove_marker_clicked), NULL);
    g_signal_connect_swapped(G_OBJECT(clear_markers), "clicked",
                             G_CALLBACK(gtk_list_store_clear), 
                             G_OBJECT(marker_list));

    GtkWidget *marker_buttons_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL,
                                                10);
    gtk_widget_set_halign(marker_buttons_box, GTK_ALIGN_CENTER);
    gtk_container_add(GTK_CONTAINER(marker_buttons_box), add_marker);
    gtk_container_add(GTK_CONTAINER(marker_buttons_box), remove_marker);
    gtk_container_add(GTK_CONTAINER(marker_buttons_box), clear_markers);

    GtkWidget *scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(GTK_CONTAINER(scrolled_window), markers);
    gtk_scrolled_window_set_policy
        (GTK_SCROLLED_WINDOW(scrolled_window),
         GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    GtkWidget *marker_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 10);
    gtk_box_pack_start(GTK_BOX(marker_box), 
                       GTK_WIDGET(scrolled_window), TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(marker_box), 
                       marker_buttons_box, FALSE, FALSE, 0);
    return marker_box;
}

/*
 * Format marker time point to two digits precision.
 */
static void
format_marker(GtkTreeViewColumn *tree_column, GtkCellRenderer *cell,
              GtkTreeModel *tree_model, GtkTreeIter *iter, gpointer data)
{
    (void)(tree_column);
    (void)(data);
    static GString *timepoint = NULL;
    if (!timepoint) timepoint = g_string_new(NULL);
    gdouble marker;
    gtk_tree_model_get(tree_model, iter, 0, &marker, -1);
    g_string_printf(timepoint, "%.2f", marker);
    GValue formatted = G_VALUE_INIT;
    g_value_init(&formatted, G_TYPE_STRING);
    g_value_set_string(&formatted, timepoint->str);
    g_object_set_property(G_OBJECT(cell), "text", &formatted);
    g_value_unset(&formatted);
}

/*
 * Create test-related widgets and lay them into a hbox.  Return the
 * box.
 */
static GtkWidget *
create_test_box(void)
{
    GtkWidget *decide = gtk_button_new_with_mnemonic("_Decide X");
    GtkWidget *test_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
    g_signal_connect(G_OBJECT(decide), "clicked",
                     G_CALLBACK(decide_button_clicked), NULL);
    trial_label = gtk_label_new(NULL);
    gtk_box_pack_start(GTK_BOX(test_box), trial_label, FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(test_box), decide, FALSE, FALSE, 0);
    return test_box;
}

/*
 * Create metadata-related widgets and lay them into a vbox.  Return
 * the box.
 */
static GtkWidget *
create_metadata_box(void)
{
    GtkWidget *sample_a_label = gtk_label_new("Sample A:");
    GtkWidget *sample_b_label = gtk_label_new("Sample B:");
    meta_a_label = gtk_label_new(NULL);
    meta_b_label = gtk_label_new(NULL);
    GtkWidget *meta_a_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    GtkWidget *meta_b_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    gtk_box_pack_start(GTK_BOX(meta_a_hbox), sample_a_label, FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(meta_a_hbox), meta_a_label, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(meta_b_hbox), sample_b_label, FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(meta_b_hbox), meta_b_label, FALSE, FALSE, 0);
    GtkWidget *meta_vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 5);
    gtk_box_pack_start(GTK_BOX(meta_vbox), meta_a_hbox, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(meta_vbox), meta_b_hbox, FALSE, FALSE, 0);
    return meta_vbox;
}

/*
 * Create menu bar and items.  Return the menu bar.
 */
static GtkWidget *
create_menu_bar(void)
{
    g_assert(main_window);      /* the main window must exist for quit callback */
    GtkWidget *test_menu = gtk_menu_new();
    new_test_menu_item = gtk_menu_item_new_with_mnemonic("_New");
    quit_menu_item = gtk_menu_item_new_with_mnemonic("_Quit");
    gtk_menu_shell_append(GTK_MENU_SHELL(test_menu), new_test_menu_item);
    gtk_menu_shell_append(GTK_MENU_SHELL(test_menu), quit_menu_item);
    g_signal_connect(G_OBJECT (new_test_menu_item), "activate",
                     G_CALLBACK(menu_item_clicked), NULL);
    g_signal_connect(G_OBJECT(quit_menu_item), "activate",
                     G_CALLBACK(menu_item_clicked), NULL);
    GtkWidget *test_menu_item = gtk_menu_item_new_with_mnemonic("_Test");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM (test_menu_item),
                              test_menu);

    GtkWidget *help_menu = gtk_menu_new();
    about_menu_item = gtk_menu_item_new_with_mnemonic("_About");
    gtk_menu_shell_append(GTK_MENU_SHELL (help_menu), about_menu_item);
    g_signal_connect(G_OBJECT (about_menu_item), "activate",
                     G_CALLBACK (menu_item_clicked), NULL);
    GtkWidget *help_menu_item = gtk_menu_item_new_with_mnemonic("_Help");
    gtk_menu_item_set_submenu(GTK_MENU_ITEM (help_menu_item),
                              help_menu);

    GtkWidget *menu_bar = gtk_menu_bar_new();
    gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), test_menu_item);
    gtk_menu_shell_append(GTK_MENU_SHELL(menu_bar), help_menu_item);
    return menu_bar;
}

/*
 * Create the status bar.
 */
static void
create_status_bar(void)
{
    status_bar = gtk_statusbar_new();
    play_context = (gtk_statusbar_get_context_id
                    (GTK_STATUSBAR(status_bar), "Playback status"));
    non_play_context = (gtk_statusbar_get_context_id
                        (GTK_STATUSBAR(status_bar), 
                         "Non-playback status"));
    gtk_statusbar_push(GTK_STATUSBAR(status_bar), non_play_context, 
                       "Stopped");
}

/*
 * Handle main window delete signal.
 *
 * Ask for confirmation if there is a test in progress.
 */
static gboolean
delete_event_handler(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    return !confirm("Are you sure you want to exit?");
}

/*
 * Handle main window destroy signal.  This will eventually exit the
 * program.
 */
static void
destroy_event_handler(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    main_window = NULL;
    gtk_main_quit();
}

/*
 * Handler marker selection (activation in GTK lingo).
 */
static void
marker_activated(GtkTreeView *markers, GtkTreePath *path,
                 GtkTreeViewColumn *column, gpointer user_data)
{
    (void)(markers);
    (void)(column);
    (void)(user_data);
    GtkTreeIter iter;
    if (!gtk_tree_model_get_iter
        (GTK_TREE_MODEL(marker_list), &iter, path)) {
        g_warning("can't get iterator for marked item");
        return;
    }
    gdouble marker;
    gtk_tree_model_get(GTK_TREE_MODEL(marker_list),
                       &iter, 0, &marker, -1);
    playback_seek(marker);
    update_location();
}

/*
 * Clear marker selection, if any.
 */
static void
clear_marker_selection(void)
{
    gtk_tree_selection_unselect_all
        (gtk_tree_view_get_selection(GTK_TREE_VIEW(markers)));
}

/*
 * Handle scale button dragging.
 */
static gboolean
scale_button_pressed_or_released(GtkWidget *widget,
                                 GdkEventButton *event,
                                 gpointer data)
{
    (void)(widget);
    (void)(data);
    static double previous = -1;
    if (event->type == GDK_BUTTON_PRESS) {
        is_user_seeking = 1;
        previous = gtk_range_get_value(GTK_RANGE(scale));
    } else if (event->type == GDK_BUTTON_RELEASE) {
        double current = gtk_range_get_value(GTK_RANGE(scale));
        if (fabs(previous - current) > 0.001) {
            clear_marker_selection();
            playback_seek(gtk_range_get_value(GTK_RANGE(scale)));
            update_location();
        }
        is_user_seeking = 0;
    }
    return FALSE;
}

static gboolean
scale_scroll_action(GtkRange *range,
                    GtkScrollType scroll,
                    double value,
                    gpointer data)
{
    (void)(scroll);
    (void)(value);
    (void)(data);
    if (!is_user_seeking) {
        double current = gtk_range_get_value(range);
        gtk_range_set_value(range, value);
        double new = gtk_range_get_value(range);
        /* Seek only when there is a difference between positions. */
        if (fabs(current - new) > 0.001) {
            clear_marker_selection();
            if (seek_delay_id) {
                g_source_remove(seek_delay_id);
            }
            seek_delay_id = g_timeout_add_once(SEEK_DELAY_TIMEOUT,
                                               delayed_seek, NULL);
        }
    }
    return FALSE;
}

/*
 * Add current location as marker.  If playback is in progress, use
 * playback location.  Otherwise use the slider value.
 */
static void
add_marker_clicked(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    if (current_trial < 0) return;
    GtkTreeIter new;
    gtk_list_store_append(marker_list, &new);
    Playback_state state = playback_get_state();
    if (state.location < 0) {
        gtk_list_store_set(marker_list, &new,
                           0, gtk_adjustment_get_value
                           (GTK_ADJUSTMENT(adjustment)),
                           -1);
    } else {
        gtk_list_store_set(marker_list, &new, 0, state.location, -1);
    }
}

/*
 * Remove the selected marker.
 */
static void
remove_marker_clicked(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    GtkTreeIter iter;
    if (gtk_tree_selection_get_selected
        (gtk_tree_view_get_selection(GTK_TREE_VIEW(markers)),
         NULL, &iter)) {
        gtk_list_store_remove(marker_list, &iter);
    }
}

/*
 * Handle menu item clicks.
 */
static void
menu_item_clicked(GtkWidget *widget, gpointer data)
{
    (void)(data);
    if (widget == new_test_menu_item) {
        if (confirm("Are you sure you want to replace the current test "
                    "with a new test?")) {
            playback_stop();
            show_new_test_window(GTK_WINDOW(main_window));
        }
    } else if (widget == about_menu_item) {
        show_about_dialog();
    } else if (widget == quit_menu_item) {
        if (confirm("Are you sure you want to exit?")) {
            gtk_widget_destroy(main_window);
        }
    } else {
        g_warning("menu item handler called with unknown widget");
    }
}

/*
 * Show decide dialog.
 */
static void
decide_button_clicked(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    if (current_trial >= 0) {
        playback_stop();
        decide_current_trial();
    }
}

/*
 * Start playback and update the status bar.
 */
static void
play_button_clicked(GtkWidget *widget, gpointer data)
{
    (void)(data);
    if (current_trial < 0) return;

    gint sampleid;
    if (widget == first_sample) {
        sampleid = 0;
    } else if (widget == second_sample) {
        sampleid = 1;
    } else {
        sampleid = test_get_answer(current_trial);
    }

    /* check for marker selection */
    GtkTreeIter iter;
    int rv;
    if (gtk_tree_selection_get_selected
        (gtk_tree_view_get_selection(GTK_TREE_VIEW(markers)),
         NULL, &iter)) {
        gdouble marker;
        gtk_tree_model_get(GTK_TREE_MODEL(marker_list),
                           &iter, 0, &marker, -1);
        rv = playback_start(sampleid, marker);
    } else {
        Playback_state state = playback_get_state();
        if (state.location < 0) {
            rv = playback_start(sampleid, gtk_adjustment_get_value
                                (GTK_ADJUSTMENT(adjustment)));
        } else {
            rv = playback_start(sampleid, state.location);
        }
    }
    if (rv != 0) {
        error_dialog(main_window, "Failed to start playback",
                     playback_error);
        return;
    }
    clear_status_bar();
    if (widget == first_sample) {
        gtk_statusbar_push(GTK_STATUSBAR(status_bar),
                           play_context, "Playing sample A");
    } else if (widget == second_sample) {
        gtk_statusbar_push(GTK_STATUSBAR(status_bar), play_context,
                           "Playing sample B");
    } else {
        gtk_statusbar_push(GTK_STATUSBAR(status_bar), play_context,
                           "Playing sample X");
    }
    update_location();
}

/*
 * Pause or resume the playback. 
 */
static void
pause_button_clicked(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    Playback_state state = playback_pause_or_resume();
    switch (state.playback) {
    case PAUSED:
        gtk_statusbar_push(GTK_STATUSBAR(status_bar), play_context, "Paused");
        break;
    case PLAYING:
        gtk_statusbar_pop(GTK_STATUSBAR(status_bar), play_context);
        break;
    default:
        clear_status_bar();
    }
    update_location();
}

/*
 * Seek to the next or previous marker.
 */
static void
rewind_button_clicked(GtkWidget *widget, gpointer data)
{
    (void)(data);
    GtkTreeSelection *selection = gtk_tree_view_get_selection
        (GTK_TREE_VIEW(markers));
    GtkTreeIter iter;
    if (gtk_tree_selection_get_selected(selection, NULL, &iter) == FALSE) {
        gtk_tree_selection_select_path(selection, gtk_tree_path_new_first());
    } else {
        GtkTreePath *path = gtk_tree_model_get_path(GTK_TREE_MODEL(marker_list),
                                                    &iter);
        if (widget == prev_marker && gtk_tree_path_prev(path)) {
            gtk_tree_selection_select_path(selection, path);
        } else if (widget == next_marker) {
            gtk_tree_path_next(path);
            gtk_tree_selection_select_path(selection, path);
        }
        gtk_tree_path_free(path);
    }
    if (gtk_tree_selection_get_selected
        (gtk_tree_view_get_selection(GTK_TREE_VIEW(markers)),
         NULL, &iter)) {
        gdouble marker;
        gtk_tree_model_get(GTK_TREE_MODEL(marker_list),
                           &iter, 0, &marker, -1);
        playback_seek(marker);
        update_location();
    }
}

/*
 * If there is a test in progress, ask for confirmation before
 * creating a new test.
 *
 * Return TRUE if affirmative or if there is not test in progress,
 * FALSE otherwise.
 */
static gboolean
confirm(const gchar *message)
{
    if (current_trial != -1) {
        GtkDialogFlags flags = GTK_DIALOG_DESTROY_WITH_PARENT;
        GtkWidget *dialog = gtk_message_dialog_new
            (GTK_WINDOW(main_window), flags,
             GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO,
             "Test in progress");
        gtk_message_dialog_format_secondary_text
            (GTK_MESSAGE_DIALOG(dialog), "%s", message);
        gtk_window_set_title(GTK_WINDOW(dialog), "Warning");
        gint result = gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        return result == GTK_RESPONSE_YES;
    } else {
        return TRUE;
    }
}

/*
 * Fill in and show the about dialog.
 */
static void
show_about_dialog(void)
{
    static const gchar *authors[] = {
        "Petteri Hintsanen <petterih@iki.fi>",
        NULL
    };
    GdkPixbuf *logo = gdk_pixbuf_new_from_resource
        ("/fi/kapsi/phintsan/abx/abx-128.png", NULL);
    gtk_show_about_dialog(GTK_WINDOW(main_window),
                          "program-name", "ABX Tester",
                          "title", "About ABX Tester",
                          "authors", authors,
                          "comments", "Fidelity testing program.",
                          "website", PACKAGE_URL,
                          "license-type", GTK_LICENSE_GPL_3_0,
                          "version", VERSION,
                          "logo", logo,
                          NULL);
}

/*
 * Show decide dialog and act according to the user response.
 */
static void
decide_current_trial(void)
{
    GtkWidget *dialog = (gtk_message_dialog_new
                         (GTK_WINDOW(main_window),
                          GTK_DIALOG_DESTROY_WITH_PARENT,
                          GTK_MESSAGE_QUESTION,
                          GTK_BUTTONS_NONE,
                          "Decide X"));
    gtk_window_set_title(GTK_WINDOW(dialog), "Decision");
    gtk_message_dialog_format_secondary_text
        (GTK_MESSAGE_DIALOG(dialog), "Which one is X, A or B?");
    gtk_dialog_add_button(GTK_DIALOG(dialog), "_A", 0);
    gtk_dialog_add_button(GTK_DIALOG(dialog), "_B", 1);
    gtk_dialog_add_button(GTK_DIALOG(dialog), "_Cancel",
                          GTK_RESPONSE_CANCEL);
    gint response = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);

    if (response == 0 || response == 1) {
        test_set_guess(current_trial, response);
        if (current_trial == test_get_ntrials() - 1) {
            /* test over */
            show_result_window(GTK_WINDOW(main_window));
        } else {
            /* proceed to the next trial */
            current_trial++;
            update_main_window();
        }
    }
}
