/* mathlib.c - library of mathematical routines
 *
 * Copyright 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mathlib.h"
#include <math.h>

double
p_value(int k, int n)
{
    if (k == 0) return 1;       /* p-value is trivially 1 */
    /* use symmetry of Bin(n, 0.5) for improved numerical accuracy */
    if (k < n / 2) {
        double F = cdf_binomial(k - 1, n, 0.5);
        return 1 - F;
    } else {
        double F = cdf_binomial(n - k, n, 0.5);
        return F;
    }
}

/*
 * Calculate cumulative distribution function value F(k) for binomial
 * distribution Bin(n, p).
 *
 * The algorithm is from Råde, Westergren: Mathematics Handbook, page
 * 426, 4th ed., 1998.  Studentlitteratur, Lund.
 */
double
cdf_binomial(int k, int n, double p)
{
    if (k >= n) return 1;
    double f = pow((1 - p), n); /* f is the pf, init to f(0) */
    double F = f;               /* F is the cdf */
    for (int x = 1; x <= k; x++) {
        f = (p / (1 - p)) * ((1.0 * n - x + 1) / x) * f;
        F += f;
    }
    return F;
}
