/* newtestwin.c - window for creating a new test
 *
 * Copyright 2010, 2016, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gtkui.h"

static GtkWidget *window;
static GtkWidget *main_box;
static GtkWidget *first_file_selector;
static GtkWidget *second_file_selector;
static GtkWidget *first_file_box;
static GtkWidget *second_file_box;
static GtkWidget *file_box;
static GtkAdjustment *adjustment;
static GtkWidget *spinner;
static GtkWidget *spinner_box;
static GtkWidget *ok;
static GtkWidget *cancel;
static GtkWidget *button_box;

/*
 * Try to initialize a new test.
 *
 * Destroy window if test setup was successful.  Otherwise display
 * error dialog.
 */
static void
init_new_test(void)
{
    gchar *a = gtk_file_chooser_get_filename
        (GTK_FILE_CHOOSER(first_file_selector));
    gchar *b = gtk_file_chooser_get_filename
        (GTK_FILE_CHOOSER(second_file_selector));

    if (a == NULL) {
        error_dialog(window, "Failed to initialize a test",
                     "You have not chosen sample file A.");
        goto out;
    }

    if (b == NULL) {
        error_dialog(window, "Failed to initialize a test",
                     "You have not chosen sample file B.");
        goto out;
    }

    if (set_up_test(a, b,
                    gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinner)),
                    window) == 0) {
        gtk_widget_destroy(window);
        reset_test_widgets();
    }

out:
    g_free(a);
    g_free(b);
}

/*
 * Handle window destroy signal.
 */
static void
destroy_event_handler(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    window = NULL;
}

/*
 * Handle button clicks.
 */
static void
button_clicked(GtkWidget *button, gpointer data)
{
    (void)(data);
    if (button == cancel) { 
        gtk_widget_destroy(window);
    } else if (button == ok) {
        init_new_test();
    }
}

/*
 * Create and show a new test window.
 */
void
show_new_test_window(GtkWindow *parent)
{
    GtkWidget *sample_a_label = gtk_label_new_with_mnemonic("Sample _A:");
    GtkWidget *sample_b_label = gtk_label_new_with_mnemonic("Sample _B:");
    GtkWidget *trials_label = (gtk_label_new_with_mnemonic
                               ("Number of _trials:"));

    /* create filename boxes and selector */
    first_file_selector = (gtk_file_chooser_button_new
                           ("Choose audio sample A", 
                            GTK_FILE_CHOOSER_ACTION_OPEN));
    second_file_selector = (gtk_file_chooser_button_new
                            ("Choose audio sample B", 
                             GTK_FILE_CHOOSER_ACTION_OPEN));
    gtk_label_set_mnemonic_widget(GTK_LABEL(sample_a_label), 
                                  first_file_selector);
    gtk_label_set_mnemonic_widget(GTK_LABEL(sample_b_label), 
                                  second_file_selector);

    first_file_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    gtk_box_pack_start(GTK_BOX(first_file_box), sample_a_label,
                       FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(first_file_box),
                     GTK_WIDGET(first_file_selector), FALSE, FALSE, 0);
    second_file_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    gtk_box_pack_start(GTK_BOX(second_file_box), sample_b_label,
                       FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(second_file_box),
                       second_file_selector, FALSE, FALSE, 0);
    file_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 10);
    gtk_box_pack_start(GTK_BOX(file_box),
                       first_file_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(file_box),
                       second_file_box, TRUE, TRUE, 0);

    /* trials */
    adjustment = gtk_adjustment_new(16, 1, 50, 1.0, 10.0, 0);
    spinner = gtk_spin_button_new(adjustment, 1, 0);
    gtk_label_set_mnemonic_widget(GTK_LABEL(trials_label),
                                  spinner);
    spinner_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    gtk_box_pack_start(GTK_BOX(spinner_box), trials_label, FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(spinner_box), spinner, FALSE, FALSE, 0);

    /* buttons */
    ok = gtk_button_new_with_mnemonic("_OK");
    cancel = gtk_button_new_with_mnemonic("_Cancel");
    button_box = gtk_button_box_new(GTK_ORIENTATION_HORIZONTAL);
    gtk_button_box_set_layout(GTK_BUTTON_BOX(button_box),
                              GTK_BUTTONBOX_CENTER);
    gtk_box_set_spacing(GTK_BOX(button_box), 10);
    gtk_container_add(GTK_CONTAINER(button_box), ok);
    gtk_container_add(GTK_CONTAINER(button_box), cancel);
    g_signal_connect(ok, "clicked", G_CALLBACK(button_clicked), NULL);
    g_signal_connect(cancel, "clicked", G_CALLBACK(button_clicked), NULL);

    /* main window */
    g_assert(window == NULL);
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
    gtk_window_set_transient_for(GTK_WINDOW(window), parent);
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_window_set_title(GTK_WINDOW(window), "New test");
    main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 20);
    gtk_container_set_border_width(GTK_CONTAINER(main_box), 10);
    gtk_box_pack_start(GTK_BOX(main_box), file_box, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), spinner_box, TRUE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), button_box, FALSE, FALSE, 0);
    g_signal_connect(window, "destroy",
                     G_CALLBACK(destroy_event_handler), NULL);
    gtk_container_add(GTK_CONTAINER(window), main_box);
    gtk_window_set_default_size(GTK_WINDOW(window), 500, 100);
    gtk_widget_show_all(window);
}
