/* playback.c - high level audio playback implementation
 *
 * Copyright 2010, 2016, 2024, 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "playback.h"
#include "player.h"
#include <glib.h>

const char *playback_error;

static Sndfile *sample_a;
static Sndfile *sample_b;
static Player *player;
static Playback_state playback_state = { STOPPED, -1 };

int
playback_init(int outdev, int latency)
{
    g_debug("initializing playback: output device %d, latency %d",
            outdev, latency);
    player = player_new(outdev, latency);
    if (!player) {
        playback_error = player_error;
        return 1;
    }
    return 0;
}

void
playback_close(void)
{
    if (!player) return;
    g_debug("closing playback");
    playback_stop();
    player_destroy(player);
}

int 
playback_open_samples(const char *a, const char *b)
{
    g_assert(sample_a == NULL);
    g_assert(sample_b == NULL);
    g_assert(a && b);
    g_debug("initializing sample playback for samples %s and %s", a, b);
    if (!(sample_a = sndfile_open(a))) {
        playback_error = sndfile_error;
        return 1;
    }
    if (!(sample_b = sndfile_open(b))) {
        playback_error = sndfile_error;
        playback_close_samples();
        return 2;
    }
    if (player_load(player, sample_a) != 0
        || player_eject(player) != 0) {
        playback_error = player_error;
        playback_close_samples();
        return 4;
    }
    if (player_load(player, sample_b) != 0
        || player_eject(player) != 0) {
        playback_error = player_error;
        playback_close_samples();
        return 5;
    }
    if (!sndfile_same_duration(sample_a, sample_b)){
        playback_error = NULL;
        return 3;
    }
    return 0;
}

void
playback_close_samples(void)
{
    if (sample_a == NULL && sample_b == NULL) return;
    g_debug("closing sample playback");
    playback_stop();
    playback_state.location = -1;
    if (sample_a && sndfile_close(sample_a) != 0) {
        g_warning("failed to close sample A: %s",
                  sndfile_error);
    }
    if (sample_b && sndfile_close(sample_b) != 0) {
        g_warning("failed to close sample B: %s",
                  sndfile_error);
    }
    sample_a = sample_b = NULL;
}

void
playback_get_metadata(Sndfile_metadata *a, Sndfile_metadata *b)
{
    g_assert(a && b);
    *a = sndfile_get_metadata(sample_a);
    *b = sndfile_get_metadata(sample_b);
}

Playback_state
playback_get_state(void)
{
    /* Player does not signal its state but has to be polled
     * instead. */
    if (player_loaded(player)) {
        /* Check if player has reached the end. */
        if (playback_state.playback == PLAYING
            && !player_playing(player)) {
            playback_state.playback = STOPPED;
        }
        playback_state.location = player_tell(player);
    }
    return playback_state;
}

int
playback_start(int sample, double pos)
{
    g_debug("start playback of sample %d from %f", sample, pos);
    Sndfile *sf;
    if (sample == 0) sf = sample_a;
    else sf = sample_b;
    Sndfile *loaded = player_loaded(player);
    /* Load the sample to player unless it is already loaded. */
    if (sf != loaded) {
        if (loaded && player_eject(player) != 0) {
            playback_error = player_error;
            return 1;
        }
        if (player_load(player, sf) != 0) {
            playback_error = player_error;
            return 2;
        }
    } else {
        player_stop(player);
    }
    if (player_seek(player, pos) == -1) {
        playback_error = player_error;
        return 3;
    }
    if (player_start(player) != 0) {
        playback_error = player_error;
        return 4;
    }
    playback_state.playback = PLAYING;
    return 0;
}

int
playback_stop(void)
{
    if (player_loaded(player) && player_stop(player) != 0) {
        playback_error = player_error;
        return 1;
    }
    playback_state.playback = STOPPED;
    return 0;
}

Playback_state
playback_pause_or_resume(void)
{
    if (player_loaded(player))
    {
        switch (playback_state.playback) {
        case PLAYING:
            if (player_stop(player) == 0) {
                playback_state.playback = PAUSED;
            } else g_warning("failed to stop playback: %s",
                             player_error);
            break;
        case PAUSED:
            if (player_start(player) == 0) {
                playback_state.playback = PLAYING;
            } else g_warning("failed to start playback: %s",
                             player_error);
            break;
        case STOPPED:
            break;
        }
    }
    return playback_state;
}

void
playback_seek(double pos)
{
    if (player_loaded(player) && player_seek(player, pos) == -1) {
        g_warning("seek to %f failed: %s", pos, player_error);
    }
}
