/* playback.h - high level audio playback interface
 *
 * Copyright 2010, 2016, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ABX_PLAYBACK_H
#define ABX_PLAYBACK_H

#include "soundfile.h"

/* Current playback state. */
typedef struct {
    enum { PLAYING, PAUSED, STOPPED } playback;
    double location;
} Playback_state;

/* Error string for the most recent error. */
extern const char *playback_error;

/*
 * Initialize playback.
 *
 * This function must be called before any other playback functions.
 * Use output device index outdev (see get_output_devices) and the
 * given output latency.
 *
 * If output device is -1, use the default output device.
 * If latency is -1, use the default output device latency.
 *
 * Return 0 on success, nonzero otherwise.
 */
extern int playback_init(int outdev, int latency);

/*
 * Close playback.
 *
 * Stop playback and free the allocated resources.
 */
extern void playback_close(void);

/*
 * Initialize playback for sample files a and b.
 *
 * The samples must have equal duration, and the player must be able
 * to load both of them.  In practice this means that the samples must
 * be in compatible format.
 *
 * This function must be called before calling playback_start or
 * playback_stop.
 *
 * Return
 *   0 on success
 *   1 if sample a could not be opened
 *   2 if sample b could not be opened
 *   3 if the samples have different duration
 *   4 if the player cannot load sample a
 *   5 if the player cannot load sample b
 */
extern int playback_open_samples(const char *a, const char *b);

/*
 * Close the current samples and free the allocated resources.
 *
 * Stop playback if it is in progress.
 */
extern void playback_close_samples(void);

/*
 * Fill the given metadata structures.
 */
extern void playback_get_metadata(Sndfile_metadata *a, Sndfile_metadata *b);

/*
 * Return the current playback state.
 */
extern Playback_state playback_get_state(void);

/*
 * Start playing sample from the given position (in seconds).
 *
 * sample must be either 0 (sample A) or nonzero (sample B).
 *
 * Return zero on success, nonzero otherwise.
 */
extern int playback_start(int sample, double pos);

/*
 * Stop playback.
 *
 * Return zero on success, nonzero otherwise.
 */
extern int playback_stop(void);

/*
 * Pause or resume playback.
 *
 * Return the playback state after change.
 */
extern Playback_state playback_pause_or_resume(void);

/*
 * Reposition playback to the given position.
 *
 * Position is in seconds, measured from the beginning of the current
 * sample.  If playback is in progress, it is continued from the new
 * position.
 */
extern void playback_seek(double pos);

#endif
