/* player.c - low level audio file player implementation
 *
 * This version uses PortAudio (http://www.portaudio.com) for audio
 * output.
 *
 * Copyright 2010, 2016, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "player.h"
#include "audio.h"
#include <glib.h>
#include <portaudio.h>
#include <stdio.h>
#include <stdlib.h>

const char *player_error;

/* Ring buffer size in milliseconds.  The actual buffer size in bytes
 * depends on sample rate and number of channels. */
static const int BUFFER_SIZE_IN_MS = 1000;

/* Minimum acceptable fill level for the buffer.  This is intended to
 * keep PortAudio stream callback from signaling spaces all the time.
 * MIN_BUFFER_LEVEL must be between 0 and 1, though not too close to
 * zero to prevent buffer underruns. */
static const float MIN_BUFFER_LEVEL = 0.5;

struct Player {
    /* Ring buffer for samples.  We always keep one vacant slot, so
     * that it is relatively easy to distinguish between empty (in ==
     * out) and full (in + 1 == out) buffer. */
    struct {
        float *begin;
        float *end;
        /* Input pointer for reader. */
        float *in;
        /* Output pointer fore stream callback. */
        float *out;
        /* Buffer size in bytes. */
        size_t size;
        /* Is the buffer fill below MIN_BUFFER_LEVEL?  This is also
         * used to signal buffer pre-fill before starting playback. */
        int space;
        GCond space_cond;
        GMutex space_mutex;
    } buffer;

    /* Reader thread data. */
    struct {
        GThread *thread;
        /* If nonzero, thread should stop as soon as possible. */
        int stop;
        /* Has reader hit end of file? */
        gint eof;
    } reader;

    /* Sample file handle. */
    Sndfile *sndfile;
    /* Sample metadata. */
    Sndfile_metadata metadata;
    /* PortAudio output device. */
    PaDeviceIndex device;
    /* Output latency in milliseconds. */
    int latency;
    /* PortAudio output stream. */
    PaStream *stream;
    /* Is the playback on? */
    int playing;
    /* Most recent seek location, in seconds. */
    double origin;
    /* Frames played after the most recent seek. */
    int nplayed;
    /* Current playback location, in seconds. */
    double location;
};

static int init_stream(Player *player, Sndfile *sf);
static int stream_callback(const void *input,
                           void *output,
                           unsigned long nframes,
                           const PaStreamCallbackTimeInfo *timeinfo,
                           PaStreamCallbackFlags statusflags,
                           void *userdata);
static void stream_finished_callback(void *userdata);
static size_t buffer_fill(Player *player);

static void start_reader(Player *player);
static void stop_reader(Player *player);
static void *reader_main(void *arg);
static int reader_prefill_buffer(Player *player);
static int reader_fill_buffer(Player *player);
static int read_frames(Player *player, size_t nframes);

Player *
player_new(int outdev, int latency)
{
    /* Scan for the default output device, or verify that the given
     * output device exists. */
    const Audio_device *devices;
    int ndevices = audio_get_outputs(&devices);
    if (ndevices == -1) return NULL;
    if (outdev == -1) {
        for (int i = 0; i < ndevices; i++) {
            if (devices[i].sysdefault) {
                outdev = i;
                break;
            }
        }
        if (outdev == -1) {
            player_error = "No default output device found";
            return NULL;
        }
    } else {
        if (outdev >= ndevices) {
            player_error = "Invalid output device";
            return NULL;
        }
    }
    Player *player = g_malloc(sizeof(Player));
    player->buffer.begin = NULL;
    player->buffer.end = NULL;
    player->buffer.in = NULL;
    player->buffer.out = NULL;
    player->buffer.size = 0;
    g_cond_init(&player->buffer.space_cond);
    g_mutex_init(&player->buffer.space_mutex);
    player->reader.thread = NULL;
    player->sndfile = NULL;
    player->device = outdev;
    /* Get the default latency or use the configured one. */
    if (latency < 0) {
        latency = devices[outdev].latency;
        if (latency > 0) {
            g_debug("using the default output latency %d ms", latency);
            player->latency = latency;
        } else {
            g_warning("failed to fetch the default output latency, "
                      "falling back to 50 ms");
            player->latency = 50;
        }
    } else {
        g_debug("using the configured latency %d ms", latency);
        player->latency = latency;
    }
    player->stream = NULL;
    g_debug("player initialized");
    return player;
}

void
player_destroy(Player *player)
{
    g_assert(player);
    g_debug("deleting player");
    g_free(player->buffer.begin);
    g_cond_clear(&player->buffer.space_cond);
    g_mutex_clear(&player->buffer.space_mutex);
    g_free(player);
}

int
player_load(Player *player, Sndfile *sf)
{
    g_assert(player && sf);
    if (player->sndfile) {
        player_error = "Player already loaded";
        return 1;
    }
    g_debug("loading file %p to player", (void *) sf);
    player->sndfile = sf;
    player->metadata = sndfile_get_metadata(sf);
    if (init_stream(player, sf) != 0) {
        player->sndfile = NULL;
        return 2;
    }
    /* Calculate buffer size and align it to the frame size. */
    size_t frame_size = player->metadata.channels * sizeof(float);
    size_t nbytes = (size_t) (BUFFER_SIZE_IN_MS
                              * (player->metadata.rate / 1000.0)
                              * frame_size
                              + 1);
    nbytes -= nbytes % frame_size;
    if (player->buffer.begin == NULL) {
        player->buffer.begin = (float *) g_malloc(nbytes);
        g_debug("allocated %lu bytes (%lu frames) for sample buffer",
                nbytes, nbytes / frame_size);
    } else if (player->buffer.size < nbytes) {
        player->buffer.begin = (float *) g_realloc(player->buffer.begin,
                                                   nbytes);
        g_debug("allocated %lu bytes (%lu frames) for sample buffer",
                nbytes, nbytes / frame_size);
    }
    player->buffer.size = nbytes;
    player->buffer.end = (player->buffer.begin + (nbytes / sizeof(float)));
    player->playing = 0;
    player->origin = 0;
    player->nplayed = 0;
    player->location = 0;
    return 0;
}

int
player_eject(Player *player)
{
    g_assert(player);
    if (!player_loaded(player)) {
        player_error = "Player not loaded";
        return 1;
    }
    g_debug("ejecting player");
    if (player_stop(player) != 0) {
        return 2;
    }
    PaError rv = Pa_CloseStream(player->stream);
    if (rv != paNoError) {
        player_error = Pa_GetErrorText(rv);
        return 3;
    }
    player->stream = NULL;
    player->sndfile = NULL;
    return 0;
}

Sndfile *
player_loaded(Player *player)
{
    g_assert(player);
    return player->sndfile;
}

int
player_start(Player *player)
{
    g_assert(player);
    if (!player_loaded(player)) {
        player_error = "Player not loaded";
        return 1;
    }
    /* Start a new reader thread and wait for it to fill the playback
     * buffer.  After that start the PortAudio stream. */
    g_debug("starting reader");
    start_reader(player);
    g_mutex_lock(&player->buffer.space_mutex);
    while (player->buffer.space) {
        g_debug("waiting for reader to fill up buffer");
        g_cond_wait(&player->buffer.space_cond,
                    &player->buffer.space_mutex);
    }
    g_mutex_unlock(&player->buffer.space_mutex);
    g_debug("buffer filled, starting playback stream");
    PaError rv = Pa_StartStream(player->stream);
    if (rv != paNoError) {
        player_error = Pa_GetErrorText(rv);
        player->playing = 0;
        stop_reader(player);
        return 1;
    }
    g_debug("playing");
    player->playing = 1;
    return 0;
}

int
player_stop(Player *player)
{
    g_assert(player);
    if (!player_loaded(player)) {
        player_error = "Player not loaded";
        return 1;
    }
    PaError rv = Pa_IsStreamStopped(player->stream);
    if (rv == 1) return 0;      /* Stream already stopped */
    if (rv < 0) {
        /* PortAudio error. */
        player_error = Pa_GetErrorText(rv);
        return 2;
    }
    g_assert(rv == 0);
    g_debug("stopping player");
    /* Pa_StopStream will cause a call to stream_finished_callback,
     * which will stop the reader and update the playback flag. */
    rv = Pa_StopStream(player->stream);
    if (rv != paNoError) {
        player_error = Pa_GetErrorText(rv);
        return 3;
    }
    return 0;
}

int
player_playing(Player *player)
{
    g_assert(player);
    if (!player_loaded(player)) return 0;
    else return player->playing;
}

double
player_seek(Player *player, double pos)
{
    g_assert(player);
    if (!player_loaded(player)) {
        player_error = "Player not loaded";
        return -1;
    }
    g_debug("seeking to %.2f seconds", pos);
    int was_playing = 0;
    if (player->playing) {
        was_playing = 1;
        if (player_stop(player) != 0) return -1;
    }
    if (pos < 0) pos = 0;
    if (pos > player->metadata.duration) pos = player->metadata.duration;
    if (sndfile_seek(player->sndfile, pos, SEEK_SET) == -1) {
        player_error = sndfile_error;
        return -1;
    }
    player->location = pos;
    player->origin = pos;
    player->nplayed = 0;
    player->buffer.in = player->buffer.out = player->buffer.begin;
    g_debug("seek done");
    if (was_playing) {
        g_debug("resuming playback");
        if (player_start(player) != 0) return -1;
    }
    return pos;
}

double
player_tell(Player *player)
{
    g_assert(player);
    if (!player_loaded(player)) return -1;
    else return player->location;
}


/*
 * Initialize PortAudio stream for playing sample file sf.
 *
 * Return zero on success, nonzero otherwise.
 */
static int
init_stream(Player *player, Sndfile *sf)
{
    g_assert(player && sf);
    g_debug("initializing playback stream");
    PaStreamParameters strparams;
    player->metadata = sndfile_get_metadata(sf);
    memset(&strparams, 0, sizeof(PaStreamParameters));
    strparams.channelCount = player->metadata.channels;
    strparams.device = player->device;
    strparams.sampleFormat = paFloat32;
    strparams.suggestedLatency = player->latency / 1000.0;
    strparams.hostApiSpecificStreamInfo = NULL;
    PaError rv = Pa_OpenStream(&player->stream, NULL, &strparams,
                               player->metadata.rate, 0,
                               paNoFlag, stream_callback, player);
    if (rv != paNoError) {
        player_error = Pa_GetErrorText(rv);
        player->stream = NULL;
        return 1;
    }
    rv = Pa_SetStreamFinishedCallback(player->stream,
                                      stream_finished_callback);
    if (rv != paNoError) {
        player_error = Pa_GetErrorText(rv);
        rv = Pa_CloseStream(player->stream);
        if (rv != paNoError) {
            g_warning("failed to close playback stream: %s",
                      Pa_GetErrorText(rv));
        }
        player->stream = NULL;
        return 2;
    }
    return 0;
}

/*
 * Stream callback for PortAudio.
 *
 * Fill the audio output buffer from the ring buffer, and signal space
 * so that the reader thread can fill the ring buffer.
 *
 * This is the consumer part of producer-consumer model.
 *
 * Return PaContinue so that PortAudio will keep on playback, or
 * PaComplete if EOF has been reached.
 */
static int
stream_callback(const void *input,
                void *output,
                unsigned long nframes,
                const PaStreamCallbackTimeInfo *timeinfo,
                PaStreamCallbackFlags statusflags,
                void *userdata)
{
    (void)(input);
    (void)(timeinfo);
    (void)(statusflags);
    Player *player = (Player *) userdata;
    int eof = 0;
    int nchannels = player->metadata.channels;
    float *outbuf = (float *) output;
    unsigned long nplayed = 0;
    while (nframes > 0) {
        if (player->buffer.out == player->buffer.end) {
            /* rewind the ring buffer */
            player->buffer.out = player->buffer.begin;
        }
        if (player->buffer.out == player->buffer.in) {
            /* Buffer underrun or EOF. */
            if (!eof) {
                gint reader_eof = g_atomic_int_get(&player->reader.eof);
                if (!reader_eof) {
                    g_warning("player: buffer underrun");
                } else {
                    g_debug("player: eof");
                    eof = 1;
                }
            }
            while (nframes > 0) {
                /* Output silence. */
                for (int i = 0; i < nchannels; i++) {
                    *outbuf++ = 0;
                }
                nframes--;
            }
        } else {
            /* Output samples from the buffer. */
            for (int i = 0; i < nchannels; i++) {
                *outbuf++ = *player->buffer.out++;
            }
            nframes--;
            nplayed++;
        }
    }
    /* Update the current playback location. */
    player->nplayed += nplayed;
    player->location = (player->origin
                        + (1.0 * player->nplayed
                           / player->metadata.rate));
    /* Signal space if the buffer is empty enough. */
    if (g_mutex_trylock(&player->buffer.space_mutex)) {
        if (buffer_fill(player) < MIN_BUFFER_LEVEL * player->buffer.size) {
            player->buffer.space = 1;
            g_cond_signal(&player->buffer.space_cond);
        }
        g_mutex_unlock(&player->buffer.space_mutex);
    }
    if (!eof) return paContinue;
    else return paComplete;
}

/*
 * Stream finished callback for PortAudio.
 *
 * Stop the reader and mark playback as stopped.
 */
static void
stream_finished_callback(void *userdata)
{
    Player *player = (Player *) userdata;
    g_assert(player && player->playing);
    g_debug("stream finished");
    player->playing = 0;
    stop_reader(player);
    g_debug("player stopped");
}

/*
 * Return the current buffer fill in bytes.
 */
static size_t
buffer_fill(Player *player)
{
    g_assert(player);
    size_t nvacant;
    float *out = player->buffer.out;
    if (player->buffer.in == out) {
        return 0;
    } else if (player->buffer.in < out) {
        nvacant = out - player->buffer.in;
    } else {
        nvacant = player->buffer.end - player->buffer.in;
        nvacant += out - player->buffer.begin;
    }
    nvacant *= sizeof(float);
    g_assert(nvacant <= player->buffer.size);
    return player->buffer.size - nvacant;
}


/*
 * Start a new reader thread for the current sound file.
 *
 * Playback must be stopped, since this function resets the buffer.
 */
static void
start_reader(Player *player)
{
    g_assert(player->sndfile);
    g_assert(!player->playing);
    g_assert(player->reader.thread == NULL);
    /* Reset the buffer and start a new reader thread. */
    player->buffer.in = player->buffer.out = player->buffer.begin;
    player->buffer.space = 1;
    player->reader.stop = 0;
    player->reader.eof = 0;
    player->reader.thread = g_thread_new("reader", reader_main, player);
}

/*
 * Terminate the reader thread.
 *
 * Playback must be stopped before calling this function.
 */
static void
stop_reader(Player *player)
{
    g_assert(!player->playing);
    if (player->reader.thread == NULL) {
        g_debug("reader not running");
        return;
    }
    g_debug("telling reader to terminate");
    g_mutex_lock(&player->buffer.space_mutex);
    player->reader.stop = 1;
    /* Signal the reader in case it is blocked. */
    player->buffer.space = 1;
    g_cond_signal(&player->buffer.space_cond);
    g_mutex_unlock(&player->buffer.space_mutex);
    g_debug("waiting for reader to terminate");
    g_thread_join(player->reader.thread);
    player->reader.thread = NULL;
    g_debug("reader has been terminated");
}

/*
 * Reader thread.
 *
 * This thread implements the producer part of producer-consumer model
 * by reading blocks of audio data from the sound file into the ring
 * buffer.
 *
 * Some notes:
 *
 * - Locks and condition variables are used for synchronization.
 *
 * - A new reader is started after each seek operation.  The reader
 *   starts reading from the sought position.
 *
 * - The reader blocks if there is no free space in the buffer.  It
 *   then waits for a space signal from the playback stream callback
 *   (consumer).
 *
 * - The reader reads from the file until EOF is hit, or until it has
 *   been explicitly requested to terminate by stop_reader.
 *
 * - The reader signals space after the buffer has been filled for the
 *   first time.  This signifies that playback can start.
 *
 * Return always NULL.
 */
static void *
reader_main(void *arg)
{
    g_debug("reader: starting");
    Player *player = (Player *) arg;
    int eof = 0;
    int prefill = 1;
    while (!eof) {
        /* Wait for space. */
        g_mutex_lock(&player->buffer.space_mutex);
        while (!player->buffer.space && !player->reader.stop) {
            /* g_debug("reader: waiting for space"); */
            g_cond_wait(&player->buffer.space_cond,
                        &player->buffer.space_mutex);
        }
        /* Check for stop request. */
        if (player->reader.stop) {
            g_debug("reader: exiting");
            g_mutex_unlock(&player->buffer.space_mutex);
            return NULL;
        }
        if (prefill) {
            eof = reader_prefill_buffer(player);
            prefill = 0;
        } else eof = reader_fill_buffer(player);
        /* The buffer has been filled. */
        player->buffer.space = 0;
        g_mutex_unlock(&player->buffer.space_mutex);
    }
    g_debug("reader: exiting");
    return NULL;
}

static int
reader_prefill_buffer(Player *player)
{
    g_assert(player->buffer.in == player->buffer.begin);
    g_assert(player->buffer.in == player->buffer.out);
    size_t frame_size = player->metadata.channels * sizeof(float);
    size_t prefill = (size_t) (MIN_BUFFER_LEVEL * player->buffer.size
                               / frame_size);
    int eof = read_frames(player, prefill);
    g_debug("reader: signaling pre-fill");
    g_cond_signal(&player->buffer.space_cond);
    return eof;
}

static int
reader_fill_buffer(Player *player)
{
    /* Take a snapshot of buffer out pointer, as the player thread
     * may adjust it concurrently. */
    float *out = player->buffer.out;
    size_t nvacant;
    int eof = 0;
    if (player->buffer.in < out) {
        /* Read until buffer out pointer, but leave one vacant
         * slot. */
        nvacant = ((out - player->buffer.in)
                   / player->metadata.channels) - 1;
        if (nvacant > 0) eof = read_frames(player, nvacant);
    } else {
        /* It is safe to read until second to last slot. */
        nvacant = ((player->buffer.end - player->buffer.in)
                   / player->metadata.channels) - 1;
        if (out > player->buffer.begin) {
            /* Since there is at least one vacant slot in the
             * beginning of the buffer, it is safe to read to the
             * end of the buffer. */
            nvacant++;
        }
        eof = read_frames(player, nvacant);
        if (player->buffer.in < out && !eof) {
            /* Fill the beginning of the buffer as well, if
             * possible. */
            nvacant = ((out - player->buffer.in)
                       / player->metadata.channels) - 1;
            if (nvacant > 0) eof = read_frames(player, nvacant);
        }
    }
    return eof;
}

/*
 * Read data from the loaded sound file.
 *
 * Read nframes frames into the player buffer.  It is assumed that the
 * loaded sound file is already at the correct position (see
 * player_seek), and that the frames will fit into the buffer.
 *
 * Rewind the ring buffer if the buffer end is reached.
 *
 * Return nonzero if EOF was hit during reading, that is, less than
 * nframes frames were read.
 */
static int
read_frames(Player *player, size_t nframes)
{
    g_assert(player && nframes > 0);
    g_assert(player->buffer.in + nframes * player->metadata.channels
             <= player->buffer.end);
    size_t nread = sndfile_read(player->sndfile, player->buffer.in, nframes);
    player->buffer.in += (nread * player->metadata.channels);
    g_assert(player->buffer.in <= player->buffer.end);
    int eof = 0;
    if (nread != nframes) {
        /* EOF */
        g_debug("reader: eof");
        g_atomic_int_set(&player->reader.eof, 1);
        eof = 1;
    }
    if (player->buffer.in == player->buffer.end) {
        /* Rewind the ring buffer. */
        player->buffer.in = player->buffer.begin;
    }
    return eof;
}
