/* player.h - low level audio player interface
 *
 * Copyright 2010, 2016, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ABX_PLAYER_H
#define ABX_PLAYER_H

#include "soundfile.h"

/* Player handle type. */
typedef struct Player Player;

/* Error string for the most recent error. */
extern const char *player_error;

/*
 * Create and return a new player.
 *
 * Use PortAudio device outdev for audio output, or scan for the
 * default host API output device if outdev is -1.
 *
 * latency is the desired output latency in milliseconds.  If it is
 * negative, try to use the default (high) output latency from
 * PortAudio.  If that fails, fall back to 50 ms.
 */
extern Player *player_new(int outdev, int latency);

/*
 * Destroy player and free the allocated resources.
 *
 * Player must be empty (ejected) before calling this function.
 */
extern void player_destroy(Player *player);

/*
 * Prepare player to play sound file sf.
 *
 * Player must be empty (ejected) before calling this function.
 *
 * Return zero on success, nonzero otherwise.
 */
extern int player_load(Player *player, Sndfile *sf);

/*
 * Eject loaded sound file from player.
 *
 * If player is playing, stop the playback.
 *
 * Return zero on success, nonzero otherwise.
 */
extern int player_eject(Player *player);

/*
 * Return the loaded sound file, or NULL if player is not loaded.
 */
extern Sndfile *player_loaded(Player *player);

/*
 * Start playback.
 *
 * Return zero on success, nonzero otherwise.
 */
extern int player_start(Player *player);

/*
 * Stop playback.
 *
 * Note that calling this function is always required before playback
 * can be restarted.  This is true even if playback has stopped
 * already due to the end of the input file.
 *
 * Return zero on success, nonzero otherwise.
 */
extern int player_stop(Player *player);

/*
 * Return nonzero if player is currently playing, or zero if it is
 * stopped.
 */
extern int player_playing(Player *player);

/*
 * Seek player to position pos in seconds.
 *
 * If pos is less than zero or larger than the duration, clamp pos to
 * a valid position.  If playback is in progress, continue it from the
 * new location.
 *
 * Return the new location if seek is successful, otherwise return -1.
 */
extern double player_seek(Player *player, double pos);

/*
 * Return the current playback position in seconds, or -1 if player is
 * not loaded.
 */
extern double player_tell(Player *player);

#endif
