/* resultwin.c - window for displaying test results
 *
 * Copyright 2010, 2013, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "test.h"
#include "gtkui.h"

static GtkWidget *answers_box;
static GtkWidget *close_box;
static GtkWidget *guesses_box;
static GtkWidget *main_box;
static GtkWidget *ncorr_box;
static GtkWidget *pval_box;
static GtkWidget *window;
static GtkWidget *close_button;
static GtkWidget *answers;
static GtkWidget *guesses;

/*
 * Handle window destroy signal.
 */
static void
destroy_event_handler(GtkWidget *widget, gpointer data)
{
    (void)(widget);
    (void)(data);
    window = NULL;
    discard_test();
    current_trial = -1;
    update_main_window();
}

/*
 * Create and show result window.
 */
void
show_result_window(GtkWindow *parent)
{
    g_assert(window == NULL);
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
    gtk_window_set_transient_for(GTK_WINDOW(window), parent);
    gtk_window_set_modal(GTK_WINDOW(window), TRUE);
    gtk_window_set_title(GTK_WINDOW(window), "Test results");
    g_signal_connect(window, "destroy",
                     G_CALLBACK(destroy_event_handler), NULL);
    close_button = gtk_button_new_with_mnemonic("_Close");
    g_signal_connect_swapped(close_button, "clicked",
                             G_CALLBACK(gtk_widget_destroy), window);

    /* build decision strings */
    GString *as = g_string_new(test_get_answer(0) == TEST_SAMPLE_A ? "<tt>A" : "<tt>B");
    GString *gs = g_string_new(test_get_guess(0) == TEST_SAMPLE_A ? "<tt>A" : "<tt>B");
    int ncorr = test_get_answer(0) == test_get_guess(0) ? 1 : 0;
    for (int i = 1; i < test_get_ntrials(); i++) {
        g_string_append(as, test_get_answer(i) == TEST_SAMPLE_A ? " A" : " B");
        g_string_append(gs, test_get_guess(i) == TEST_SAMPLE_A ? " A" : " B");
        if (test_get_answer(i) == test_get_guess(i)) ncorr++;
    }
    g_string_append(as, "</tt>");
    g_string_append(gs, "</tt>");
    answers = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL (answers), as->str);
    guesses = gtk_label_new(NULL);
    gtk_label_set_markup(GTK_LABEL (guesses), gs->str);
    g_string_free(as, TRUE);
    g_string_free(gs, TRUE);

    answers_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    guesses_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    ncorr_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    pval_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);
    close_box = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 10);

    gtk_box_pack_start(GTK_BOX(answers_box), 
                       gtk_label_new("Correct decisions:"), 
                       FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(answers_box), answers, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX(guesses_box), 
                       gtk_label_new("Your decisions:"),
                       FALSE, FALSE, 0);
    gtk_box_pack_end(GTK_BOX(guesses_box), guesses, FALSE, FALSE, 0);

    gtk_box_pack_start(GTK_BOX(ncorr_box),
                       gtk_label_new("Number of correct decisions:"),
                       FALSE, FALSE, 0);
    gchar *cs = g_strdup_printf("%d", ncorr);
    gtk_box_pack_end(GTK_BOX(ncorr_box),
                     gtk_label_new(cs),
                     FALSE, FALSE, 0);
    g_free(cs);

    gtk_box_pack_start(GTK_BOX(pval_box),
                       gtk_label_new("p-value:"),
                       FALSE, FALSE, 0);
    gchar *pvals = g_strdup_printf("%.3g", test_get_pvalue());
    gtk_box_pack_end(GTK_BOX(pval_box),
                     gtk_label_new(pvals),
                     FALSE, FALSE, 0);
    g_free(pvals);

    gtk_box_pack_start(GTK_BOX(close_box),
                       close_button, TRUE, FALSE, 0);

    main_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 10);
    gtk_container_set_border_width(GTK_CONTAINER(main_box), 10);
    gtk_box_pack_start(GTK_BOX(main_box), answers_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), guesses_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), ncorr_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), pval_box, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(main_box), close_box, TRUE, TRUE, 0);

    gtk_container_add(GTK_CONTAINER(window), main_box);
    gtk_widget_show_all(window);
}
