/* soundfile.c - low level sound file I/O implementation
 *
 * Based on libsndfile (http://www.mega-nerd.com/libsndfile/).
 *
 * Copyright 2010, 2016, 2024, 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "soundfile.h"
#include <glib.h>
#include <sndfile.h>
#include <math.h>

struct Sndfile {
    SNDFILE *sffile;
    Sndfile_metadata metadata;
};

const char *sndfile_error = NULL;

Sndfile *
sndfile_open(const char *filename)
{
    g_assert(filename);
    g_debug("opening sound file %s", filename);
    SF_INFO sfinfo;
    sfinfo.format = 0;
    SNDFILE *sffile = sf_open(filename, SFM_READ, &sfinfo);
    if (!sffile) {
        sndfile_error = sf_strerror(NULL);
        return NULL;
    }
    Sndfile *sndfile = g_malloc(sizeof(Sndfile));
    sndfile->sffile = sffile;
    sndfile->metadata.filename = g_strdup(filename);
    sndfile->metadata.rate = sfinfo.samplerate;
    sndfile->metadata.channels = sfinfo.channels;
    sndfile->metadata.frames = sfinfo.frames;
    sndfile->metadata.duration = (1.0 * sndfile->metadata.frames 
                                  / sndfile->metadata.rate);
    return sndfile;
}

int
sndfile_close(Sndfile *sndfile)
{
    g_assert(sndfile);
    g_debug("closing sound file %s", sndfile->metadata.filename);
    int rv = sf_close(sndfile->sffile);
    if (rv != 0) sndfile_error = sf_error_number(rv);
    g_free(sndfile->metadata.filename);
    g_free(sndfile);
    return rv;
}

Sndfile_metadata
sndfile_get_metadata(Sndfile *sndfile)
{
    g_assert(sndfile);
    return sndfile->metadata;
}

double
sndfile_seek(Sndfile *sndfile, double offset, int whence)
{
    g_assert(sndfile);
    sf_count_t frames = sndfile->metadata.rate * offset;
    g_debug("seeking to frame %ld / %ld", frames, sndfile->metadata.frames);
    sf_count_t loc = sf_seek(sndfile->sffile, frames, whence);
    if (loc != -1) loc = 1.0 * loc / sndfile->metadata.rate;
    else sndfile_error = sf_strerror(sndfile->sffile);
    return loc;
}

long
sndfile_read(Sndfile *sndfile, float *buf, long nframes)
{
    g_assert(sndfile && buf);
    return sf_readf_float(sndfile->sffile, buf, nframes);
}

int
sndfile_same_duration(Sndfile *a, Sndfile *b)
{
    g_assert(a && b);
    return fabs(a->metadata.duration - b->metadata.duration) <= 0.01;
}
