/* soundfile.h - low level sound file I/O interface
 *
 * Copyright 2010, 2024, 2025 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ABX_SOUNDFILE_H
#define ABX_SOUNDFILE_H

#include <stddef.h>

/* Sound metadata structure. */
typedef struct {
    char *filename;
    int channels;               /* number of channels */
    long frames;                /* number of frames */
    int rate;                   /* sampling rate */
    double duration;            /* duration in seconds */
} Sndfile_metadata;

/* Sound file handle type. */
typedef struct Sndfile Sndfile;

/* Error string for the most recent error. */
extern const char *sndfile_error;

/*
 * Open sound file for reading.
 *
 * Return handle to the opened sound file, or NULL on error.
 */
extern Sndfile *sndfile_open(const char *filename);

/*
 * Close sound file.
 *
 * Return 0 on success, or nonzero on error.
 */
extern int sndfile_close(Sndfile *sndfile);

/*
 * Return sound file metadata.
 */
extern Sndfile_metadata sndfile_get_metadata(Sndfile *sndfile);

/*
 * Seek sound file.  Offset (in seconds) and whence are as in fseek.
 *
 * Return the new location in seconds, or -1 on error.
 */
extern double sndfile_seek(Sndfile *sndfile, double offset, int whence);

/*
 * Read nframes frames of normalized PCM data from sound file to buf.
 *
 * Return the number of frames read.  It can be less than nframes if
 * the end of file was reached.
 */
extern long sndfile_read(Sndfile *sndfile, float *buf, long nframes);

/*
 * Return nonzero if the given files have approximately the same
 * duration, otherwise zero.
 *
 * Up to 10 ms difference is allowed.
 */
extern int sndfile_same_duration(Sndfile *a, Sndfile *b);

#endif
