/* test.c - ABX test implementation
 *
 * Copyright 2010, 2016, 2024 Petteri Hintsanen <petterih@iki.fi>
 *
 * This file is part of abx.
 *
 * abx is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * abx is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 * License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with abx.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "test.h"
#include "mathlib.h"
#include <glib.h>

/*
 * Ongoing test structure.
 */
typedef struct {
    int isvalid;     /* nonzero if test is valid */
    int ntrials;     /* number of trials, must be greater than zero */
    Test_answer *answers; /* correct answers array with ntrials elements */
    Test_answer *guesses;  /* guesses array with ntrials elements */
} Test;

static Test current_test = { 0, -1, NULL, NULL };

int
test_init(int ntrials)
{
    if (ntrials == 0) return -1;
    test_close();
    current_test.guesses = g_malloc(ntrials * sizeof(Test_answer));
    current_test.answers = g_malloc(ntrials * sizeof(Test_answer));
    for (int i = 0; i < ntrials; i++) {
        if (g_random_boolean() == TRUE) current_test.answers[i] = TEST_SAMPLE_A;
        else current_test.answers[i] = TEST_SAMPLE_B;
        current_test.guesses[i] = TEST_SAMPLE_NA;
    }
    current_test.ntrials = ntrials;
    current_test.isvalid = 1;
    return 0;
}

void
test_close(void)
{
    g_free(current_test.guesses);
    g_free(current_test.answers);
    current_test.answers = current_test.guesses = NULL;
    current_test.ntrials = -1;
    current_test.isvalid = 0;
}

Test_answer
test_get_guess(int t)
{
    if (current_test.isvalid && t >= 0 && t < current_test.ntrials) {
        return current_test.guesses[t];
    } else return -1;
}

Test_answer
test_get_answer(int t)
{
    if (current_test.isvalid && t >= 0 && t < current_test.ntrials) {
        return current_test.answers[t];
    } else return -1;
}

int
test_set_guess(int t, Test_answer ans)
{
    if (current_test.isvalid && t >= 0 && t < current_test.ntrials) {
        current_test.guesses[t] = ans;
        return 0;
    } else return -1;
}

int
test_get_ntrials(void)
{
    if (current_test.isvalid) return current_test.ntrials;
    else return -1;
}

double
test_get_pvalue(void)
{
    int k = 0;                  /* the number of correct guesses */
    int n = test_get_ntrials();
    for (int i = 0; i < n; i++) {
        if (test_get_answer(i) == test_get_guess(i)) k++;
    }
    return p_value(k, n);
}
